
#include "GibbsSampler.h"
#include <cmath>
#include <cstdlib>
#include <ctime>
#include <stdexcept>

extern "C"  {
#include "linsubs.h"
#include "eigsubs.h" 
};




//note, this is duplicated in the main file
const int CONTROL_STATUS = 1;
const int CASE_STATUS = 2;


void GibbsSampler::printTime(char* message)
{
   // current date/time based on current system
   //http://www.tutorialspoint.com/cplusplus/cpp_date_time.htm
   time_t now = time(0);
   tm *ltm = localtime(&now);
   // print various components of tm structure.
   cout<<message<<endl;
   cout << "Month/Day: "<< 1 + ltm->tm_mon<< "/ "<<  ltm->tm_mday << endl;
   cout << "Time: "<< 1 + ltm->tm_hour << ":";
   cout << 1 + ltm->tm_min << ":";
   cout << 1 + ltm->tm_sec << endl;
     
}


/*
Just prints a matrix out to the console.
Should put this in some sort of static class for generic matrix stuff.
*/
void GibbsSampler::printMatrix(gsl_matrix *theMatrix, int numRow, int numCol)
{
     for(int  row = 0; row < numRow; row++){
            cout<<"[";
            for(int  col = 0; col < numCol; col++){
                           cout<<gsl_matrix_get(theMatrix,row,col);
                           if(col!=numCol-1) {cout<<", ";}
            }
            cout<<" ]"<<endl;
     }     
}


/*
Just prints a matrix out to the console.
Should put this in some sort of static class for generic matrix stuff.
*/
void GibbsSampler::printMatrix(double *theMatrix, int numRow, int numCol)
{
     for(int  row = 0; row < numRow; row++){
            cout<<"[";
            for(int  col = 0; col < numCol; col++){
                           cout<<theMatrix[row*numRow+col];
                           if(col!=numCol-1) {cout<<", ";}
            }
            cout<<" ]"<<endl;
     }     
}


/*
Just prints a matrix out to the console.
Should put this in some sort of static class for generic matrix stuff.
*/
void GibbsSampler::printMatrixDiag(gsl_matrix *theMatrix, int numRow)
{
     for(int  row = 0; row < numRow; row++){
            cout<<"[";
            cout<<gsl_matrix_get(theMatrix,row,row);
            if(row!=numRow-1) {cout<<", ";}
     }     
     cout<<" ]"<<endl;
}




void GibbsSampler::runmcmc(gsl_matrix *S, int *P, double threshold, int nindiv,int niter, int nburnin, gsl_vector* outvec, bool notRaoBlack)
{
  //need to probably pass in T and P (cov and case control status)    
  //also the threshold
  //may need to pass in nindiv
  //  srand(time(NULL));
  
  //copying matrix for test output

  
  gsl_matrix* copyPhenoCovariance = gsl_matrix_alloc(nindiv,nindiv);
  gsl_matrix_memcpy(copyPhenoCovariance,S);
  
  
  printTime("Inverting pheno cov inside of runmcmc");
  //now we'll get the inverse of our covariance matrix: Tinv    
  invertT3(S, nindiv);
  printTime("Done inverting pheno cov inside of runmcmc");
  //printMatrixDiag(S, nindiv);
  //ioh->writeInverse(nindiv, Tinv);

  //TODO remove this, just testing
  cout<<"Diagonal component of Tinv after invertint T: "<<endl;
  for(int i=0;i<nindiv;i++)  {
          cout<<Tinv[i*nindiv+i]<<",";
  }
  cout<<endl;

  //what are each of these different phi values? the means? not sure we need all
  double *phi = new double[nindiv];
  double *newphi = new double[nindiv];
  double *rbphi = new double[nindiv];
  double *phisum = new double[nindiv];
  double *rbphisum = new double[nindiv];
  
  double *phiOld = new double[nindiv];

 

  for(int i=0;i<nindiv;i++)  {
    //Setting initial values for indiiduals
    phi[i] = expTruncNormal(P[i], 0.0, 1.0, threshold);

    if (phi[i] > 10 || phi[i] < -10 || isnan(phi[i]) ){
      double alternativePhi = sampleTruncNormal(P[i], 0.0, 1.0,threshold);
      cout<<"WARNING: Marginal PML appears unreasonable: "<< phi[i] << " at index " << i << "that is a ";
      if(P[i] == CASE_STATUS){
        cout<<"Case"<<endl;
      }
      else{
        cout<<"Control"<<endl;
      }
      cout<<"An alternative might be using another method: " << alternativePhi << endl;

    }

    //if(P[i] == CONTROL_STATUS){
//        phi[i] = -0.5;         
//    }
        // phi[i] = sampleTruncNormal(P[i], 0.0, 1.0);
    phisum[i] = 0.0;
    rbphisum[i] = 0.0;
  }

    printTime("performing some validation check on the inverse of the covariance matrix");
    try{
      // check Tinv 
      //performing some validation check on the inverse of the covariance matrix
      for(int j=0;j<nindiv;j++)  {
        double sum = 0.0;
        for(int k=0;k<nindiv;k++)  {
          if ( k != j )  {
            sum += Tinv[j*nindiv+k]*phi[k];
          }
        }
        double pmean = -sum/Tinv[j*nindiv+j];
        double pvar  = 1.0/Tinv[j*nindiv+j];
    
        if( isnan(Tinv[j*nindiv+j]) ){
            cout << "error: nan diagonal element in row " << j << " of inverse of covariance matrix" << endl;
            throw std::out_of_range("error: nan diagonal element in row of inverse of covariance matrix");
        }  

        if ( pvar < 0.0 )  {
          cout << "error: negative diagonal element in row " << j << " of inverse of covariance matrix" << endl;
          cout<<"The start of the phenotypic covariance matrix is: "<<endl;
          printMatrix(copyPhenoCovariance,5,5);
          throw std::out_of_range("error: negative diagonal element in row of inverse of covariance matrix");
          //exit(1);
        }
        if ( P[j] == 1 && pmean + 10*sqrt(pvar) < threshold )  {
          cout<<endl<<endl;
          cout << "error: impossible mean and variance in row " << j << " of inverse of covariance matrix" << endl;
          cout << "  mean = " << pmean << " variance = " << pvar << endl;
          //TODO: Should put back in just testing for now
          //throw std::out_of_range("error: impossible mean and variance in row of inverse of covariance matrix");
    //      exit(1);
        }
      }
    
      printTime("Done performing some validation check on the inverse of the covariance matrix, burn in next");
    }
    catch(std::out_of_range &e)
    {
             cout<<"Error in attempting to validate within runmcmc"<<endl;              
             cout<<"Exception caught: "<< e.what() <<endl;                    
    }
    
  //keeping track of the last time the values exploded
  int lastIterResetPML = 0;  
  for(int i=0;i<niter+nburnin;i++)  {

    double sum;
    for(int j=0;j<nindiv;j++)  {

      sum = 0.0;
      for(int k=0;k<nindiv;k++)  {
        if ( k != j )  {
          sum += Tinv[j*nindiv+k] * phi[k];
          
          //---------------Testing ---------------//
          if(isnan(phi[k]) || isnan(Tinv[j*nindiv+k]) || isnan(sum) ){
            cout<<endl<<endl;
            cout<<"Issue inside of summation phi[k]: "<<phi[k]<<" for i,j,k :"<< i<<","<<j<<","<<k<<endl;
            cout<<"sum and Tinv[j*nindiv+k]: "<<sum<<", "<<Tinv[j*nindiv+k]<<endl;
            cout<<"phi[k-1] and Tinv[j*nindiv+k-1]: "<< phi[k-1] <<", "<< Tinv[j*nindiv+k-1] <<endl;
            double alternativePhi = sampleTruncNormal(P[k], 0.0, 1.0,threshold);
            cout<<"An alternative might be using another method: " << alternativePhi << endl;
            throw std::out_of_range("error: sum is coming out nan");
            //---------------Testing ---------------//
          }  

        }
      }
      
      //what are pmean and psigma?
      double pmean = -sum/Tinv[j*nindiv+j];
      double psigma  = 1.0/sqrt(Tinv[j*nindiv+j]);      
      //May want to put this back in

      //-------------Testing ----------------------//
      /*if(pmean > 100 || pmean < -100 || isnan(Tinv[j*nindiv+j])  ||  isnan(sum) ||  isnan(pmean) || isnan(newphi[j]) || isnan(psigma))
      {
          cout<<"values going to explode: "<< endl;
          cout<<"sum and Tinv[j*nindiv+j]: "<<sum<<", "<<Tinv[j*nindiv+j]<<endl;
          cout<<"indices,i,j,phi,pmean,psigma, and pheno: "<<i<<", "<<j<<", "<<newphi[j]<<", "<<pmean<<", "<<psigma<<", "<<P[j]<<endl;
          //throw std::out_of_range("pmean out of range");
      }*/ 
      //-------------Testing ----------------------//

      //TODO double check this, it appears that there is an issue when psigma
      //or pmean being is high.
      //The truncated normal appears invalid values. So, trying with
      //the other method. I'm guessing I wrote one and Sam wrote the other.
      //newphi[j] = sampleTruncNormal(P[j], pmean, psigma, threshold);
      newphi[j] = expTruncNormal(P[j], pmean, psigma, threshold);
      int sampleAttempts = 0;
      while( ((newphi[j] * newphi[j]) > 100 || isnan(newphi[j])  ) && sampleAttempts<100){
                          
           newphi[j] = expTruncNormal(P[j], pmean, psigma, threshold);
           sampleAttempts++;
      }
      
      //To avoid exploiding, reset to the univariate PML
      if((newphi[j] * newphi[j]) > 100 || isnan(newphi[j]) ){
           newphi[j] = expTruncNormal(P[j], 0, 1.0, threshold);
           lastIterResetPML = i;
      }

      
      if(newphi[j]>100 || newphi[j] < -100)
      {
          cout<<"indices,i,j,phi,pmean,psigma, and pheno: "<<i<<", "<<j<<", "<<newphi[j]<<", "<<pmean<<", "<<psigma<<", "<<P[j]<<endl;
//          throw std::out_of_range("phi,pmean,psigma, or pheno out of randge");
      }
      
      rbphi[j] = expTruncNormal(P[j], pmean, psigma, threshold);

      //To avoid exploiding, reset to the univariate PML
      //TODO, may need to remove this, if value explodes resetting it to marginal
      if((rbphi[j] * rbphi[j]) > 100 || isnan(rbphi[j]) ){      
           rbphi[j] = expTruncNormal(P[j], 0.0, 1.0, threshold);
           lastIterResetPML = i;
      }

    }

    // update phi and mean
    for(int l=0;l<nindiv;l++)  {
      phiOld[l] = phi[l];
      phi[l] = newphi[l];
      //TODO remove this
    }
    
    /*
    cout << "Iter = " << i << endl;
    cout << "  Phi    = [ ";
    for(int j=0;j<nindiv;j++)  {
      cout << setw(14) << setprecision(4) << newphi[j]; 
    }
    cout << "]\n";
    */ 

    //only store after the burn in
    if ( i < nburnin )  {
      continue;
    }
    // update phisum and meansum
    for(int l=0;l<nindiv;l++)  {
      phisum[l] += phi[l];
      rbphisum[l] += rbphi[l];
    }

  } 


  //double *outvec = new double[nindiv];
  //outvec = gsl_vector_alloc(nindiv);
  //cout<<"Before finalizing: "<<endl;
//  cout<<"Using rao blackwell: "<<!notRaoBlack<<endl;
//  cout<<"family size: "<<nindiv<<endl;
  if(notRaoBlack)
  {
//       cout<<"Final random numbers: ";
       for(int i=0;i<nindiv;i++)  {
        double  pmlFam = phi[i];
        if(i == nindiv - 1){pmlFam = phiOld[i];}
        if(abs(pmlFam)> 10 || isnan(pmlFam) ){
          cout<<"Warning, PML of "<<pmlFam<<" for individual"<<i<<endl;
        }
        gsl_vector_set(outvec, i, pmlFam);
 //       cout<<phiOld[i] <<", "<<phi[i]<<" : ";
      }
      //cout<<endl;      
//      for(int i=0;i<nindiv;i++)  {
//              cout<<gsl_vector_get(outvec, i)<<", ";
//      }
//      cout<<endl;
//      system("PAUSE");    
  }
  else
  {
      for(int i=0;i<nindiv;i++)  {
        double  pmlFam = rbphisum[i]/niter;
        if(abs(pmlFam)> 10){
          cout<<"Warning, PML of "<<pmlFam<<" for individual"<<i<<endl;
        }
        gsl_vector_set(outvec, i, pmlFam);

      }
  }
  //outvec appears to be what I want, the PML for the family members
  //ioh->writeOutput(nindiv, outvec);
  //delete outvec;
//  cout<<"about to free up at end of runmcmc"<<endl;
//  cout<<"phi[0] = "<<phi[0]<<endl;
  
  



  cout<<"The last iteration reset to the  univariate PML: " <<lastIterResetPML<<" out of "<<niter+nburnin<<endl;
  
  if(phi){delete phi;}
  if(newphi){delete newphi;}
  if(phisum){delete phisum;}
//  phi = NULL;
//  newphi = NULL;
//  phisum = NULL;
  //outvec is a pointer to the PML for the family member's and will be the
  //end result we want
  

  
  gsl_matrix_free(copyPhenoCovariance);
    printTime("Done with runmcmc");

}


//Simple method added in multiplying the liablity by the inverse phenotypic
//covariance matrix (Tinv)
void GibbsSampler::calcPsuedoPheno(int nindiv, gsl_vector* liability, gsl_vector* pseudoPheno)
{
 
 //this gets a little confusing because we're multiplying a gsl vector and a
 //standard array pointer, so we have to loop through the columns ad rows
 //individually
    for(int j=0;j<nindiv;j++)  {
        double sum = 0.0;
        for(int k=0;k<nindiv;k++)  {
            sum += Tinv[j*nindiv+k]*  gsl_vector_get(liability, k);
        }
        gsl_vector_set(pseudoPheno,j, sum);
    }
}

/*
This function converts the heritability on the liability scale to the observed
scale.
*/
double GibbsSampler::convertHeritLiabToObserv(double h2L, double threshold)
{
    //1 - disease prevalence
    double prev =  gsl_cdf_gaussian_P(threshold, 1);
    //z score
    double z =  exp(-threshold*threshold)/sqrt(2* M_PI);
    double h2Obsv = (h2L * z*z)/(prev*(1-prev));    
    return (h2Obsv);
}

/*
calculates the pseudo phenotype on the observed scale
inverse(observCov)x PML indiv = pseudoPhenoIndiv
passing by reference the pseudoPhenoIndiv
*/
void GibbsSampler::calcPsuedoPhenoObserv(gsl_matrix* observCov,
      int numFamilyMembers, gsl_vector* pmlIndiv, gsl_vector* pseudoPhenoIndiv, 
      bool testing)
{
     //invert(observCov)   
      int signum;
      int stat;
    
      if(testing)
      {
                 cout<<"Inside of calcPsuedoPhenoObserv: "<<endl;
                 printMatrix(observCov, numFamilyMembers, numFamilyMembers);
      }

    
      gsl_matrix* invObservCov = gsl_matrix_alloc(numFamilyMembers, numFamilyMembers);     // will store inverse
      gsl_permutation *p = gsl_permutation_alloc(numFamilyMembers);
    
      gsl_permutation_init(p);
      stat = gsl_linalg_LU_decomp(observCov,p,&signum);
      if ( stat )  {
        cout << "error: " << gsl_strerror(stat) << endl;
        exit(1);
      }
    
      stat = gsl_linalg_LU_invert(observCov,p,invObservCov);
      if ( stat )  {
        cout << "error: " << gsl_strerror(stat) << endl;
        exit(1);
      }
      
      //done inverting
     //multiply 
     //observCov^-1 x pmlIndiv = pseudoPhenoIndiv
     gsl_blas_dgemv(CblasNoTrans, 1, invObservCov, pmlIndiv, 0, pseudoPhenoIndiv); 

      if(testing){
                 cout<<"matrix inverted: "<<endl;
                 printMatrix(invObservCov, numFamilyMembers, numFamilyMembers);
      }


      gsl_matrix_free(invObservCov);
      gsl_permutation_free(p);

     
}



#define MAX 100.0
#define R2PIINV 0.389423
double GibbsSampler::sampleTruncNormal(int pheno, double mu, double sigma, double threshold)  {
//need to add threshold to arguments
  double a, b; 
  if ( pheno == CASE_STATUS)  {
    a = (threshold-mu)/sigma; 
    b = MAX;
  }
  else if(pheno == CONTROL_STATUS) {
    a = -MAX;
    b = (threshold-mu)/sigma; 
  }
  else  
  {
        //sometimes we won't have case control status, so do not perform 
        //a probit transformation 
       a = -MAX;
       b = MAX; 
  }
  

  double phia = gsl_cdf_gaussian_P(a,1.0);
  double phib = gsl_cdf_gaussian_P(b,1.0);

  double u =0;
  while(u==0 || u==1)
  {
    u = (double) rand()/RAND_MAX;
  }

  double v = gsl_cdf_gaussian_Pinv(phia + u*(phib-phia), 1.0);
  
  
  return(mu + sigma*v);
}


void GibbsSampler::invertT(gsl_matrix *S, int nindiv)  
{
  int signum;
  int stat;

  //gsl_matrix *S = gsl_matrix_alloc(nindiv, nindiv);     // will store LU decomp
  gsl_matrix *SI = gsl_matrix_alloc(nindiv, nindiv);     // will store inverse
  gsl_permutation *p = gsl_permutation_alloc(nindiv);

//  for(int i=0;i<nindiv;i++)  {
//    for(int j=0;j<nindiv;j++)  {
//      gsl_matrix_set(S,i,j,T[i*nindiv+j]);
//    }
//  }

  gsl_permutation_init(p);
  stat = gsl_linalg_LU_decomp(S,p,&signum);
  if ( stat )  {
    cout << "error: " << gsl_strerror(stat) << endl;
    exit(1);
  }

  stat = gsl_linalg_LU_invert(S,p,SI);
  if ( stat )  {
    cout << "error: " << gsl_strerror(stat) << endl;
    exit(1);
  }


  Tinv = new double[nindiv*nindiv];
  for(int i=0;i<nindiv;i++)  {
    for(int j=0;j<nindiv;j++)  {
      Tinv[i*nindiv+j] = gsl_matrix_get(SI,i,j);
    }
  }


  gsl_matrix_free(S); 
  gsl_matrix_free(SI); 
  gsl_permutation_free(p);
  //return Tinv;
}


double GibbsSampler::expTruncNormal(int pheno, double mu, double sigma, double threshold)  {
//need to add threshold to arguments
  double x = (threshold - mu)/sigma;
  double smallphi = gsl_ran_gaussian_pdf(x,1.0);
  double largephi = gsl_cdf_gaussian_P(x,1.0);

  //TODO, there is an issue with converging, this is exploding so trying
  //to put in a huristic way of scaling)
  //Should probably take this out
//  if(largephi<.00001){
//     cout<<"largephi was extremely small, Warning, resetting it, maybe issues with convergence"<<largephi<<endl;
//     largephi = .0001;
//  }
//  if(largephi>.99999){
//     cout<<"largephi was extremely large, Warning, resetting it, maybe issues with convergence"<<largephi<<endl;                     ;
//     largephi = .999;
//  }

  if ( pheno == CASE_STATUS)  {   // case
    return ( mu + sigma*smallphi/(1.0-largephi) );
  }
  else if(pheno == CONTROL_STATUS) 
    {           // control
    return ( mu - sigma*smallphi/largephi );
  }
  else
  {
  
      //we don't have case control status, so the mean is just the input mean  
      return (mu);


  }
}


void GibbsSampler::invertT2(gsl_matrix *S, int nindiv)
{

  cout<<"Inside invertT2"<<endl;
  double *SS = new double[nindiv*nindiv];
  cout<<"Initiated SS"<<endl;
  //TODO may need to put this back in but looks like this causes problems
//  if ( Tinv != NULL )  {
//    free(Tinv);
//  }
  Tinv = new double[nindiv*nindiv];
  cout<<"Initiated Tinv"<<endl;
  /* copy S into an array of doubles and zero out SI */
  for(int i=0;i<nindiv;i++)  {
  for(int j=0;j<nindiv;j++)  {
    SS[i*nindiv+j] = gsl_matrix_get(S,i,j);
    Tinv[i*nindiv+j] = 0.0;
  }
  }
  cout<<"About to run pdinv"<<endl;
  pdinv(Tinv,SS,nindiv);
  cout<<"finished pdinv"<<endl;
  
  cout<<"Diagonal component of Tinv before exit invertT2: "<<endl;
  for(int i=0;i<nindiv;i++)  {
          cout<<Tinv[i*nindiv+i]<<",";
  }
  cout<<endl;
  
  if ( SS != NULL ){
    free(SS);
  }
  cout<<"finished invertT2"<<endl;

}

// invert using LAPACK LU decomposition with EIGENSOFT C-wrapper
void GibbsSampler::invertT3(gsl_matrix *S, int nindiv)
{
     
//May want to put this check back in but I think it's maybe causing problems
//  if ( Tinv != NULL )  {
//    free(Tinv);
//  }
  Tinv = new double[nindiv*nindiv];

  cout << "using LAPACK invert routine\n";

  for(int i=0;i<nindiv;i++)  {
  for(int j=0;j<nindiv;j++)  {
    Tinv[i*nindiv+j] = gsl_matrix_get(S,i,j);
  }
  }
  inverse(Tinv,nindiv);

}

// invert using LAPACK Cholesky decomposition with EIGENSOFT C-wrapper
void GibbsSampler::invertT4(gsl_matrix *S, int nindiv)
{

//May want to put this check back in but I think it's maybe causing problems
//  if ( Tinv != NULL )  {
//    free(Tinv);
//  }
  Tinv = new double[nindiv*nindiv];

  cout << "using LAPACK pdinvert routine\n";

  double epsilon = 0.0;
  for(int i=0;i<nindiv;i++)  {
    for(int j=0;j<nindiv;j++)  {
      Tinv[i*nindiv+j] = gsl_matrix_get(S,i,j);
    }
    epsilon += gsl_matrix_get(S,i,i);   // trace(S)
  }

  // Cf. email from Nick, 3/13/2014, 4:58 p.m.
  epsilon /= (double)nindiv;
  epsilon *= 0.0001;

  for(int i=0;i<nindiv;i++)  {
    Tinv[i*nindiv+i] += epsilon;
  }

  inverse(Tinv,nindiv);

}


//------------Testing ------------//
  /*
  cout<<"Setting Marginal phi: "<<endl;
  cout<<"Case/Control, initialPhi, alt Phi"<<endl;
  */
  //------------Testing ------------//
    //------------Testing ------------//
    /*
    if(P[i] == CASE_STATUS){
        cout<<"Case";
    }
    else{
        cout<<"Control";
    }
    cout<<", "<<phi[i];
    double alternativePhi = sampleTruncNormal(P[i], 0.0, 1.0,threshold);
    cout<<", " << alternativePhi << endl;
    */
    //------------Testing ------------//

/*            cout<<endl<<endl;
            cout<<"Running at j: "<<j<<endl;
            cout<<"newK, Tinv[j*nindiv+newK], phi[newK]";
            for(int newK =0; newK<k; newK++){
              cout<<newK <<", "<<Tinv[j*nindiv+newK] <<", "<< phi[newK]<<endl;
            }
            cout<<endl<<endl;
*/
